@@ -33,13 +33,12 @@
 #ifdef HAVE_SPEECH
 #include <QTextToSpeech>
 #endif
+#include <QSettings>
 
 #include <sonnet/configdialog.h>
 #include <sonnet/dialog.h>
 #include <sonnet/backgroundchecker.h>
 #include <sonnet/spellcheckdecorator.h>
-#include <kconfig.h>
-#include <kconfiggroup.h>
 #include <kcursor.h>
 #include <kstandardaction.h>
 #include <kicontheme.h>
@@ -62,27 +61,27 @@ private:
     KTextEdit *m_textEdit;
 };
 
-class KTextEdit::Private
+class Q_DECL_HIDDEN KTextEdit::Private
 {
 public:
     Private(KTextEdit *_parent)
         : parent(_parent),
+          languagesMenu(nullptr),
           customPalette(false),
-          checkSpellingEnabled(false),
+          spellCheckingEnabled(false),
           findReplaceEnabled(true),
           showTabAction(true),
           showAutoCorrectionButton(false),
-          decorator(0), findDlg(0), find(0), repDlg(0), replace(0),
+          decorator(nullptr), speller(nullptr), findDlg(nullptr), find(nullptr), repDlg(nullptr), replace(nullptr),
 #ifdef HAVE_SPEECH
-          textToSpeech(Q_NULLPTR),
+          textToSpeech(nullptr),
 #endif
           findIndex(0), repIndex(0),
           lastReplacedPosition(-1)
     {
         //Check the default sonnet settings to see if spellchecking should be enabled.
-        KConfig sonnetKConfig(QLatin1String("sonnetrc"));
-        KConfigGroup group(&sonnetKConfig, "Spelling");
-        checkSpellingEnabled = group.readEntry("checkerEnabledByDefault", false);
+        QSettings settings(QStringLiteral("KDE"), QStringLiteral("Sonnet"));
+        spellCheckingEnabled = settings.value(QStringLiteral("checkerEnabledByDefault"), false).toBool();
 #ifdef HAVE_SPEECH
         textToSpeech = new QTextToSpeech(parent);
 #endif
@@ -95,6 +94,7 @@ public:
         delete find;
         delete replace;
         delete repDlg;
+        delete speller;
 #ifdef HAVE_SPEECH
         delete textToSpeech;
 #endif
@@ -137,15 +137,17 @@ public:
     QAction *autoSpellCheckAction;
     QAction *allowTab;
     QAction *spellCheckAction;
+    QMenu *languagesMenu;
     bool customPalette : 1;
 
-    bool checkSpellingEnabled : 1;
+    bool spellCheckingEnabled : 1;
     bool findReplaceEnabled: 1;
     bool showTabAction: 1;
     bool showAutoCorrectionButton: 1;
     QTextDocumentFragment originalDoc;
     QString spellCheckingLanguage;
     Sonnet::SpellCheckDecorator *decorator;
+    Sonnet::Speller *speller;
     KFindDialog *findDlg;
     KFind *find;
     KReplaceDialog *repDlg;
@@ -172,7 +174,7 @@ void KTextEdit::Private::checkSpelling(bool force)
         backgroundSpellCheck->changeLanguage(spellCheckingLanguage);
     }
     Sonnet::Dialog *spellDialog = new Sonnet::Dialog(
-        backgroundSpellCheck, force ? parent : 0);
+        backgroundSpellCheck, force ? parent : nullptr);
     backgroundSpellCheck->setParent(spellDialog);
     spellDialog->setAttribute(Qt::WA_DeleteOnClose, true);
     spellDialog->activeAutoCorrect(showAutoCorrectionButton);
@@ -180,8 +182,8 @@ void KTextEdit::Private::checkSpelling(bool force)
             parent, SLOT(spellCheckerCorrected(QString,int,QString)));
     connect(spellDialog, SIGNAL(misspelling(QString,int)),
             parent, SLOT(spellCheckerMisspelling(QString,int)));
-    connect(spellDialog, SIGNAL(autoCorrect(QString,QString)),
-            parent, SLOT(spellCheckerAutoCorrect(QString,QString)));
+    connect(spellDialog, &Sonnet::Dialog::autoCorrect,
+            parent, &KTextEdit::spellCheckerAutoCorrect);
     connect(spellDialog, SIGNAL(done(QString)),
             parent, SLOT(spellCheckerFinished()));
     connect(spellDialog, SIGNAL(cancel()),
@@ -191,13 +193,13 @@ void KTextEdit::Private::checkSpelling(bool force)
     connect(spellDialog, SIGNAL(stop()),
             parent, SLOT(spellCheckerFinished()));
     */
-    connect(spellDialog, SIGNAL(spellCheckStatus(QString)),
-            parent, SIGNAL(spellCheckStatus(QString)));
-    connect(spellDialog, SIGNAL(languageChanged(QString)),
-            parent, SIGNAL(languageChanged(QString)));
+    connect(spellDialog, &Sonnet::Dialog::spellCheckStatus,
+            parent, &KTextEdit::spellCheckStatus);
+    connect(spellDialog, &Sonnet::Dialog::languageChanged,
+            parent, &KTextEdit::languageChanged);
     if (force) {
         connect(spellDialog, SIGNAL(done(QString)), parent, SIGNAL(spellCheckingFinished()));
-        connect(spellDialog, SIGNAL(cancel()), parent, SIGNAL(spellCheckingCanceled()));
+        connect(spellDialog, &Sonnet::Dialog::cancel, parent, &KTextEdit::spellCheckingCanceled);
         //Laurent in sonnet/dialog.cpp we emit done(QString) too => it calls here twice spellCheckerFinished not necessary
         //connect(spellDialog, SIGNAL(stop()), parent, SIGNAL(spellCheckingFinished()));
     }
@@ -307,8 +309,8 @@ void KTextEdit::Private::slotReplaceText(const QString &text, int replacementInd
 void KTextEdit::Private::init()
 {
     KCursor::setAutoHideCursor(parent, true, false);
-    parent->connect(parent, SIGNAL(languageChanged(QString)),
-                    parent, SLOT(setSpellCheckingLanguage(QString)));
+    parent->connect(parent, &KTextEdit::languageChanged,
+                    parent, &KTextEdit::setSpellCheckingLanguage);
 }
 
 KTextDecorator::KTextDecorator(KTextEdit *textEdit):
@@ -364,7 +366,7 @@ void KTextEdit::showSpellConfigDialog(const QString &windowIcon)
         dialog.setLanguage(d->spellCheckingLanguage);
     }
     if (!windowIcon.isEmpty()) {
-        dialog.setWindowIcon(QIcon::fromTheme(windowIcon));
+        dialog.setWindowIcon(QIcon::fromTheme(windowIcon, dialog.windowIcon()));
     }
     if (dialog.exec()) {
         setSpellCheckingLanguage(dialog.language());
@@ -527,7 +529,7 @@ QMenu *KTextEdit::mousePopupMenu()
 {
     QMenu *popup = createStandardContextMenu();
     if (!popup) {
-        return 0;
+        return nullptr;
     }
     connect(popup, SIGNAL(triggered(QAction*)),
             this, SLOT(menuActivated(QAction*)));
@@ -536,7 +538,7 @@ QMenu *KTextEdit::mousePopupMenu()
     if (!isReadOnly()) {
         QList<QAction *> actionList = popup->actions();
         enum { UndoAct, RedoAct, CutAct, CopyAct, PasteAct, ClearAct, SelectAllAct, NCountActs };
-        QAction *separatorAction = 0L;
+        QAction *separatorAction = nullptr;
         int idx = actionList.indexOf(actionList[SelectAllAct]) + 1;
         if (idx < actionList.count()) {
             separatorAction = actionList.at(idx);
@@ -555,15 +557,45 @@ QMenu *KTextEdit::mousePopupMenu()
 
     if (!isReadOnly()) {
         popup->addSeparator();
-        d->spellCheckAction = popup->addAction(QIcon::fromTheme(QLatin1String("tools-check-spelling")),
-                                               i18n("Check Spelling..."));
-        if (emptyDocument) {
-            d->spellCheckAction->setEnabled(false);
+        if (!d->speller) {
+            d->speller = new Sonnet::Speller();
+        }
+        if (!d->speller->availableBackends().isEmpty()) {
+
+            d->spellCheckAction = popup->addAction(QIcon::fromTheme(QStringLiteral("tools-check-spelling")),
+                                                   i18n("Check Spelling..."));
+            if (emptyDocument) {
+                d->spellCheckAction->setEnabled(false);
+            }
+            if (checkSpellingEnabled()) {
+                d->languagesMenu = new QMenu(i18n("Spell Checking Language"), popup);
+                QActionGroup *languagesGroup = new QActionGroup(d->languagesMenu);
+                languagesGroup->setExclusive(true);
+
+                QMapIterator<QString, QString> i(d->speller->availableDictionaries());
+                const QString language = spellCheckingLanguage();
+                while (i.hasNext()) {
+                    i.next();
+
+                    QAction *languageAction = d->languagesMenu->addAction(i.key());
+                    languageAction->setCheckable(true);
+                    languageAction->setChecked(language == i.value() || (language.isEmpty()
+                                                                         && d->speller->defaultLanguage() == i.value()));
+                    languageAction->setData(i.value());
+                    languageAction->setActionGroup(languagesGroup);
+                    connect(languageAction, &QAction::triggered,
+                            [this, languageAction]() {
+                        setSpellCheckingLanguage(languageAction->data().toString());
+                    });
+                }
+                popup->addMenu(d->languagesMenu);
+            }
+
+            d->autoSpellCheckAction = popup->addAction(i18n("Auto Spell Check"));
+            d->autoSpellCheckAction->setCheckable(true);
+            d->autoSpellCheckAction->setChecked(checkSpellingEnabled());
+            popup->addSeparator();
         }
-        d->autoSpellCheckAction = popup->addAction(i18n("Auto Spell Check"));
-        d->autoSpellCheckAction->setCheckable(true);
-        d->autoSpellCheckAction->setChecked(checkSpellingEnabled());
-        popup->addSeparator();
         if (d->showTabAction) {
             d->allowTab = popup->addAction(i18n("Allow Tabulations"));
             d->allowTab->setCheckable(true);
@@ -580,8 +612,8 @@ QMenu *KTextEdit::mousePopupMenu()
             findNextAction->setEnabled(false);
             findPrevAction->setEnabled(false);
         } else {
-            findNextAction->setEnabled(d->find != 0);
-            findPrevAction->setEnabled(d->find != 0);
+            findNextAction->setEnabled(d->find != nullptr);
+            findPrevAction->setEnabled(d->find != nullptr);
         }
         popup->addSeparator();
         popup->addAction(findAction);
@@ -599,9 +631,9 @@ QMenu *KTextEdit::mousePopupMenu()
 #ifdef HAVE_SPEECH
     popup->addSeparator();
     QAction *speakAction = popup->addAction(i18n("Speak Text"));
-    speakAction->setIcon(QIcon::fromTheme(QLatin1String("preferences-desktop-text-to-speech")));
+    speakAction->setIcon(QIcon::fromTheme(QStringLiteral("preferences-desktop-text-to-speech")));
     speakAction->setEnabled(!emptyDocument);
-    connect(speakAction, SIGNAL(triggered(bool)), this, SLOT(slotSpeakText()));
+    connect(speakAction, &QAction::triggered, this, &KTextEdit::slotSpeakText);
 #endif
     return popup;
 }
@@ -639,14 +671,14 @@ Sonnet::Highlighter *KTextEdit::highlighter() const
     if (d->decorator) {
         return d->decorator->highlighter();
     } else {
-        return 0;
+        return nullptr;
     }
 }
 
 void KTextEdit::clearDecorator()
 {
     delete d->decorator;
-    d->decorator = 0;
+    d->decorator = nullptr;
 }
 
 void KTextEdit::addTextDecorator(Sonnet::SpellCheckDecorator *decorator)
@@ -657,6 +689,8 @@ void KTextEdit::addTextDecorator(Sonnet::SpellCheckDecorator *decorator)
 void KTextEdit::setHighlighter(Sonnet::Highlighter *_highLighter)
 {
     KTextDecorator *decorator = new KTextDecorator(this);
+    // The old default highlighter must be manually deleted.
+    delete decorator->highlighter();
     decorator->setHighlighter(_highLighter);
 
     //KTextEdit used to take ownership of the highlighter, Sonnet::SpellCheckDecorator does not.
@@ -668,7 +702,7 @@ void KTextEdit::setHighlighter(Sonnet::Highlighter *_highLighter)
 void KTextEdit::setCheckSpellingEnabled(bool check)
 {
     emit checkSpellingChanged(check);
-    if (check == d->checkSpellingEnabled) {
+    if (check == d->spellCheckingEnabled) {
         return;
     }
 
@@ -676,7 +710,7 @@ void KTextEdit::setCheckSpellingEnabled(bool check)
     // on that we need to create a new highlighter and if we're turning it
     // off we should remove the old one.
 
-    d->checkSpellingEnabled = check;
+    d->spellCheckingEnabled = check;
     if (check) {
         if (hasFocus()) {
             createHighlighter();
@@ -691,7 +725,7 @@ void KTextEdit::setCheckSpellingEnabled(bool check)
 
 void KTextEdit::focusInEvent(QFocusEvent *event)
 {
-    if (d->checkSpellingEnabled && !isReadOnly() && !d->decorator) {
+    if (d->spellCheckingEnabled && !isReadOnly() && !d->decorator) {
         createHighlighter();
     }
 
@@ -700,7 +734,7 @@ void KTextEdit::focusInEvent(QFocusEvent *event)
 
 bool KTextEdit::checkSpellingEnabled() const
 {
-    return d->checkSpellingEnabled;
+    return d->spellCheckingEnabled;
 }
 
 bool KTextEdit::shouldBlockBeSpellChecked(const QString &) const
@@ -710,7 +744,7 @@ bool KTextEdit::shouldBlockBeSpellChecked(const QString &) const
 
 void KTextEdit::setReadOnly(bool readOnly)
 {
-    if (!readOnly && hasFocus() && d->checkSpellingEnabled && !d->decorator) {
+    if (!readOnly && hasFocus() && d->spellCheckingEnabled && !d->decorator) {
         createHighlighter();
     }
 
@@ -720,7 +754,7 @@ void KTextEdit::setReadOnly(bool readOnly)
 
     if (readOnly) {
         delete d->decorator;
-        d->decorator = 0;
+        d->decorator = nullptr;
 
         d->customPalette = testAttribute(Qt::WA_SetPalette);
         QPalette p = palette();
@@ -773,7 +807,7 @@ void KTextEdit::replace()
     } else {
         d->repDlg = new KReplaceDialog(this, 0,
                                        QStringList(), QStringList(), false);
-        connect(d->repDlg, SIGNAL(okClicked()), this, SLOT(slotDoReplace()));
+        connect(d->repDlg, &KFindDialog::okClicked, this, &KTextEdit::slotDoReplace);
     }
     d->repDlg->show();
 }
@@ -787,7 +821,7 @@ void KTextEdit::slotDoReplace()
 
     if (d->repDlg->pattern().isEmpty()) {
         delete d->replace;
-        d->replace = 0;
+        d->replace = nullptr;
         ensureCursorVisible();
         return;
     }
@@ -803,7 +837,7 @@ void KTextEdit::slotDoReplace()
     // of found text.
     connect(d->replace, SIGNAL(highlight(QString,int,int)),
             this, SLOT(slotFindHighlight(QString,int,int)));
-    connect(d->replace, SIGNAL(findNext()), this, SLOT(slotReplaceNext()));
+    connect(d->replace, &KFind::findNext, this, &KTextEdit::slotReplaceNext);
     connect(d->replace, SIGNAL(replace(QString,int,int,int)),
             this, SLOT(slotReplaceText(QString,int,int,int)));
 
@@ -846,7 +880,7 @@ void KTextEdit::slotReplaceNext()
         d->replace->displayFinalDialog();
         d->replace->disconnect(this);
         d->replace->deleteLater(); // we are in a slot connected to m_replace, don't delete it right away
-        d->replace = 0;
+        d->replace = nullptr;
         ensureCursorVisible();
         //or           if ( m_replace->shouldRestart() ) { reinit (w/o FromCursor) and call slotReplaceNext(); }
     } else {
@@ -862,7 +896,7 @@ void KTextEdit::slotDoFind()
     }
     if (d->findDlg->pattern().isEmpty()) {
         delete d->find;
-        d->find = 0;
+        d->find = nullptr;
         return;
     }
     delete d->find;
@@ -876,7 +910,7 @@ void KTextEdit::slotDoFind()
     // of found text.
     connect(d->find, SIGNAL(highlight(QString,int,int)),
             this, SLOT(slotFindHighlight(QString,int,int)));
-    connect(d->find, SIGNAL(findNext()), this, SLOT(slotFindNext()));
+    connect(d->find, &KFind::findNext, this, &KTextEdit::slotFindNext);
 
     d->findDlg->close();
     d->find->closeFindNextDialog();
@@ -891,7 +925,7 @@ void KTextEdit::slotFindNext()
     if (document()->isEmpty()) {
         d->find->disconnect(this);
         d->find->deleteLater(); // we are in a slot connected to m_find, don't delete right away
-        d->find = 0;
+        d->find = nullptr;
         return;
     }
 
@@ -905,7 +939,7 @@ void KTextEdit::slotFindNext()
         d->find->displayFinalDialog();
         d->find->disconnect(this);
         d->find->deleteLater(); // we are in a slot connected to m_find, don't delete right away
-        d->find = 0;
+        d->find = nullptr;
         //or           if ( m_find->shouldRestart() ) { reinit (w/o FromCursor) and call slotFindNext(); }
     } else {
         //m_find->closeFindNextDialog();
@@ -935,7 +969,7 @@ void KTextEdit::slotFind()
         KWindowSystem::activateWindow(d->findDlg->winId());
     } else {
         d->findDlg = new KFindDialog(this);
-        connect(d->findDlg, SIGNAL(okClicked()), this, SLOT(slotDoFind()));
+        connect(d->findDlg, &KFindDialog::okClicked, this, &KTextEdit::slotDoFind);
     }
     d->findDlg->show();
 }
@@ -951,7 +985,7 @@ void KTextEdit::slotReplace()
     } else {
         d->repDlg = new KReplaceDialog(this, 0,
                                        QStringList(), QStringList(), false);
-        connect(d->repDlg, SIGNAL(okClicked()), this, SLOT(slotDoReplace()));
+        connect(d->repDlg, &KFindDialog::okClicked, this, &KTextEdit::slotDoReplace);
     }
     d->repDlg->show();
 }
